#!/bin/sh

 ###########################################################################
 ##                                                                       ##
 ##                Centre for Speech Technology Research                  ##
 ##                     University of Edinburgh, UK                       ##
 ##                         Copyright (c) 1996                            ##
 ##                        All Rights Reserved.                           ##
 ##                                                                       ##
 ##  Permission is hereby granted, free of charge, to use and distribute  ##
 ##  this software and its documentation without restriction, including   ##
 ##  without limitation the rights to use, copy, modify, merge, publish,  ##
 ##  distribute, sublicense, and/or sell copies of this work, and to      ##
 ##  permit persons to whom this work is furnished to do so, subject to   ##
 ##  the following conditions:                                            ##
 ##   1. The code must retain the above copyright notice, this list of    ##
 ##      conditions and the following disclaimer.                         ##
 ##   2. Any modifications must be clearly marked as such.                ##
 ##   3. Original authors' names are not deleted.                         ##
 ##   4. The authors' names are not used to endorse or promote products   ##
 ##      derived from this software without specific prior written        ##
 ##      permission.                                                      ##
 ##                                                                       ##
 ##  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        ##
 ##  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      ##
 ##  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   ##
 ##  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     ##
 ##  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    ##
 ##  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   ##
 ##  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          ##
 ##  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       ##
 ##  THIS SOFTWARE.                                                       ##
 ##                                                                       ##
 ###########################################################################
 ##                                                                       ##
 ##                 Author: Richard Caley (rjc@cstr.ed.ac.uk)             ##
 ## --------------------------------------------------------------------  ##
 ## Create a build tree shadowing a number of source trees.		  ##
 ##                                                                       ##
 ###########################################################################

rm=/bin/rm
new=false

if [ "$1" = "-new" ]
    then
    new=true
    shift
fi

build=${1:?'Build Tree Name Needed'}

shift

build_below() {
    (
    root=$1
    build=$2
    source=$3
    [ -d $build ] || mkdir $build

    for s in $source/*
	do
	n=`basename $s`
	case "$n" in
	*~ ) : ;;
	#* ) : ;;
	make.* ) : ;;
	CVS ) : ;;
	* ) if [ -f $s ]
		then
		if [ -h $build/$n ] 
		    then
		    $new && [ $n != Makefile ] &&  echo "$s used"
		    $rm $build/$n
		fi
		ln -s $root/$s $build/$n
	    elif [ -d $s ]
		then
		build_below ../$root $build/$n $source/$n
	    fi;;
	esac
    done
    )
}

$new && $rm -rf "$build"

[ -d $build ] || new=true

root=..

for source in "$@"
    do
    build_below $root "$build" "$source"
done
exit 0;
