 /************************************************************************/
 /*                                                                      */
 /*                Centre for Speech Technology Research                 */
 /*                     University of Edinburgh, UK                      */
 /*                       Copyright (c) 1996,1997                        */
 /*                        All Rights Reserved.                          */
 /*                                                                      */
 /*  Permission is hereby granted, free of charge, to use and distribute */
 /*  this software and its documentation without restriction, including  */
 /*  without limitation the rights to use, copy, modify, merge, publish, */
 /*  distribute, sublicense, and/or sell copies of this work, and to     */
 /*  permit persons to whom this work is furnished to do so, subject to  */
 /*  the following conditions:                                           */
 /*   1. The code must retain the above copyright notice, this list of   */
 /*      conditions and the following disclaimer.                        */
 /*   2. Any modifications must be clearly marked as such.               */
 /*   3. Original authors' names are not deleted.                        */
 /*   4. The authors' names are not used to endorse or promote products  */
 /*      derived from this software without specific prior written       */
 /*      permission.                                                     */
 /*                                                                      */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK       */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING     */
 /*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT  */
 /*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE    */
 /*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   */
 /*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN  */
 /*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,         */
 /*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF      */
 /*  THIS SOFTWARE.                                                      */
 /*                                                                      */
 /*************************************************************************/
 /*                                                                       */
 /*                 Author: Richard Caley (rjc@cstr.ed.ac.uk)             */
 /* --------------------------------------------------------------------  */
 /* Auxiliary functions related to items.                                 */
 /*                                                                       */
 /*************************************************************************/

#include "ling_class/EST_item_aux.h"
#include "ling_class/EST_Item.h"
#include "EST_String.h"
#include "EST_error.h"

#include "../base_class/EST_get_function_template.h"

defineGetFunction(EST_Item, f, EST_Val, getVal)
defineGetFunction(EST_Item, f, EST_String, getString)
defineGetFunction(EST_Item, f, float, getFloat)
defineGetFunction(EST_Item, f, int, getInteger) 

float start(const EST_Item &item)
{
    float v=0;
    EST_feat_status status=efs_ok;
    EST_Item * pp;

    v = getFloat(item, "start", -1.0, status);

    if (v<0.0)
      {
	if ((pp = item.prev()) != NULL)
	  v = getFloat(*pp, "end", -1.0, status);
      }

    return v;
}

float mid(const EST_Item &item)
{
    float v = 0;
    EST_feat_status status=efs_ok;

    v = getFloat(item, "mid", -1.0, status);

    if (v<0.0)
      v = (start(item)+end(item))/2.0;

    return v;
}

float time(const EST_Item &item)
{

  float v = 0;
  EST_feat_status status=efs_ok;

  v = getFloat(item, "time", -1.0, status);
  
  if (v<0.0)
    v = mid(item);
  
  return v;
}

float end(const EST_Item &item)
{
  float v=0;
  EST_feat_status status=efs_ok;
  EST_Item * nn;
  
  v = getFloat(item, "end", -1.0, status);
  
  if (v < 0.0)
    {
      if ((nn = item.next()) != NULL)
	v = getFloat(*nn, "start", -1.0, status);
    }
  
  return v;
}

