#!/bin/sh
###########################################################################
##                                                                       ##
##                Centre for Speech Technology Research                  ##
##                     University of Edinburgh, UK                       ##
##                       Copyright (c) 1996,1997                         ##
##                        All Rights Reserved.                           ##
##                                                                       ##
##  Permission is hereby granted, free of charge, to use and distribute  ##
##  this software and its documentation without restriction, including   ##
##  without limitation the rights to use, copy, modify, merge, publish,  ##
##  distribute, sublicense, and/or sell copies of this work, and to      ##
##  permit persons to whom this work is furnished to do so, subject to   ##
##  the following conditions:                                            ##
##   1. The code must retain the above copyright notice, this list of    ##
##      conditions and the following disclaimer.                         ##
##   2. Any modifications must be clearly marked as such.                ##
##   3. Original authors' names are not deleted.                         ##
##   4. The authors' names are not used to endorse or promote products   ##
##      derived from this software without specific prior written        ##
##      permission.                                                      ##
##                                                                       ##
##  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        ##
##  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      ##
##  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   ##
##  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     ##
##  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    ##
##  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   ##
##  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          ##
##  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       ##
##  THIS SOFTWARE.                                                       ##
##                                                                       ##
###########################################################################
#
#  Make (a guess at the) wagon description file from a data file
#  and a list of field names
#
#  Usage: make_desc DATAFILE FIELDSFILE DESCFILE
#

if [ $# != 3 ]
then
   echo "Usage: make_desc DATAFILE FIELDSFILE DESCFILE"
   exit 1
fi

rm -f $3
echo "(" >$3
fnum=1
for i in `cat $2`
do
   echo "(" $i >> $3
   awk '{n=$'$fnum'
         if (n ~ /,/)
            printf("\"%s\"\n",n)
         else if (n ~ /'"'"'/)
            printf("\"%s\"\n",n)
         else if (n ~ /"/)
            printf("\"\\%s\"\n",n)
         else
            printf("%s\n",n)}' $1 | 
   sort -u | fold -s >> $3
   echo ")" >>$3
   fnum=`expr $fnum + 1`
done
echo ")" >>$3

exit 0




