/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                      Copyright (c) 1996,1997                          */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission is hereby granted, free of charge, to use and distribute  */
/*  this software and its documentation without restriction, including   */
/*  without limitation the rights to use, copy, modify, merge, publish,  */
/*  distribute, sublicense, and/or sell copies of this work, and to      */
/*  permit persons to whom this work is furnished to do so, subject to   */
/*  the following conditions:                                            */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*   4. The authors' names are not used to endorse or promote products   */
/*      derived from this software without specific prior written        */
/*      permission.                                                      */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*                     Author :  Alan W Black                            */
/*                     Date   :  October 1997                            */
/*-----------------------------------------------------------------------*/
/*  Some high level functions used to interpret CART trees               */
/*                                                                       */
/*=======================================================================*/
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cstring>
#include "EST_cutils.h"
#include "EST_Token.h"
#include "EST_Wagon.h"

int wagon_ask_question(LISP question, LISP value)
{
    const char *str_oper = wgn_ques_oper_str(question);

    if (streq("is",str_oper))
    {
	if (streq(get_c_string(value),
		  get_c_string(wgn_ques_operand(question))))
	    return TRUE;
	else
	    return FALSE;
    }
    else if (streq("=",str_oper))
	if (get_c_float(value) == get_c_float(wgn_ques_operand(question)))
	    return TRUE;
	else
	    return FALSE;
    else if (streq("<",str_oper))
	if (get_c_float(value) < get_c_float(wgn_ques_operand(question)))
	    return TRUE;
	else
	    return FALSE;
    else if (streq(">",str_oper))
	if (get_c_float(value) > get_c_float(wgn_ques_operand(question)))
	    return TRUE;
	else
	    return FALSE;
    else if (streq("matches",str_oper))
    {
	EST_Regex rgx(get_c_string(wgn_ques_operand(question)));
	if (EST_String(get_c_string(value)).matches(rgx))
	    return TRUE;
	else
	    return FALSE;
    }
    else if (streq("in",str_oper))
	if (siod_member_str(get_c_string(value),
			    wgn_ques_operand(question)) != NIL)
	    return TRUE;
	else
	    return FALSE;
    else 
    {
	cerr << "WAGON: unknown question operator: \"" << 
	    str_oper << "\"\n";
	siod_error();
    }
    // not reached, just for the compiler
    return FALSE;
}
